import pandas as pd
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.metadata.SignalMetadata import SignalMetadata

def get_lhc_parameters_from_pm_response(response, parameters):
    parameter_list = []
    for entry in response['content'][0]['namesAndValues']:
        if entry['name'] == 'registerNames':
            parameter_list = entry['value']

    parameter_to_value = {}
    for entry in response['content'][0]['namesAndValues']:
        if entry['name'] == 'registerValues':
            for parameter in parameters:
                index = parameter_list.index(parameter)
                parameter_to_value[parameter] = entry['value'][index]

    timestamp = response['content'][0]['rawDataEntry']['pmEventStamp']

    return pd.DataFrame(parameter_to_value, index=[timestamp])


def get_lbds_parameters_from_pm_response(response, parameters):
    parameter_to_value = {}
    for entry in response['content'][0]['namesAndValues']:
        for parameter in parameters:
            if entry['name'] == parameter:
                parameter_to_value[parameter] = max(entry['value'])

    # check for signal validity
    for entry in response['content'][0]['namesAndValues']:
        if (entry['name'] == 'dataValid') and (entry['value'] == 'False'):
            parameter_to_value = {parameter: 0.0 for parameter in parameters}

    timestamp = response['content'][0]['rawDataEntry']['pmEventStamp']

    return pd.DataFrame(parameter_to_value, index=[timestamp])


def get_blm_parameters_from_pm_response(response, parameters):
    parameter_to_value = {}
    for entry in response['content'][0]['namesAndValues']:
        for parameter in parameters:
            if entry['name'] == parameter:
                parameter_to_value[parameter] = entry['value']

    timestamp = response['content'][0]['rawDataEntry']['pmEventStamp']

    if all([not isinstance(value, (list, tuple)) for value in parameter_to_value.values()]):
        return pd.DataFrame(parameter_to_value, index=[timestamp])
    else:
        blm_df = pd.DataFrame(parameter_to_value)
        blm_df['timestamp'] = timestamp
        return blm_df


def get_blm_signal(response, blm_lst, signal, bit_convertion=4e-05):
    blm_to_signal = {}
    for entry in response['content'][0]['namesAndValues']:
        if entry['name'] == signal:
            for blm_lst_el in blm_lst:
                index = blm_lst.index(blm_lst_el)
                blm_to_signal[blm_lst_el] = entry['value'][index]

    blm_df = pd.DataFrame(blm_to_signal)
    new_index = np.arange(0, len(blm_df), 1) * bit_convertion
    blm_df.set_index(new_index, inplace=True)

    return blm_df


def convert_table_df_into_row(df, index):
    df_out = pd.DataFrame(df.stack())
    df_out.index = df_out.index.map('_'.join)
    df_out = df_out.T
    df_out.set_index(index, inplace=True)

    return df_out


def get_features_pm_blm(df, features, prefix):
    df_out = df.agg(features).T
    df_out.columns = ['%s_%s' % (prefix, col) for col in df_out.columns]

    return df_out


def get_lhc_context(start_time_date):
    t_start, t_end = Time.get_query_period_in_unix_time(start_time_date=start_time_date,
                                                        duration_date=[(1, 's'), (2, 's')])

    # Fill number
    response_blm = PmDbRequest.get_response("pmdata", False, True, pm_rest_api_path="http://pm-api-pro/v2/",
                                            system='BLM', className='BLMLHC', source='HC.BLM.SR6.C',
                                            fromTimestampInNanos=t_start, durationInNanos=int(2e9))
    pm_fill_num_df = get_blm_parameters_from_pm_response(response_blm, parameters=["pmFillNum"])

    # Beam energy and intensity
    response_lhc_a = PmDbRequest.get_response("pmdata", False, True, pm_rest_api_path="http://pm-api-pro/v2/",
                                              system='LHC', className='CISX', source='CISX.CCR.LHC.A',
                                              fromTimestampInNanos=t_start, durationInNanos=int(2e9))
    lhc_parameters_df = get_lhc_parameters_from_pm_response(response_lhc_a,
                                                            parameters=["OVERALL_ENERGY", "OVERALL_INTENSITY_1",
                                                                        "OVERALL_INTENSITY_2"])
    lhc_parameters_df['OVERALL_ENERGY'] *= 1.2  # magic number

    # Beam Mode
    response_lhc_ga = PmDbRequest.get_response("pmdata", False, True, pm_rest_api_path="http://pm-api-pro/v2/",
                                               system='LHC', className='CISX', source='CISX.CCR.LHC.GA',
                                               fromTimestampInNanos=t_start, durationInNanos=int(2e9))

    lhc_bm_df = get_lhc_parameters_from_pm_response(response_lhc_ga, parameters=["BEAM_MODE"])
    beam_mode_df = SignalMetadata.get_beam_mode_details()
    lhc_bm_df['BEAM_MODE'] = beam_mode_df[beam_mode_df['Mode'] == lhc_bm_df['BEAM_MODE'].values[0]]['Name'].values[0]

    # Abort gap
    response_lbds_b1 = PmDbRequest.get_response("pmdata", False, True, pm_rest_api_path="http://pm-api-pro/v2/",
                                                system='LBDS', className='BSRA', source='LHC.BSRA.US45.B1',
                                                fromTimestampInNanos=t_start, durationInNanos=int(5e9))
    abort_gap_b1_df = get_lbds_parameters_from_pm_response(response_lbds_b1, parameters=['aGXpocTotalIntensity',
                                                                                         'aGXpocTotalMaxIntensity'])
    abort_gap_b1_df.columns = ['%sB1' % col for col in abort_gap_b1_df.columns]

    response_lbds_b1 = PmDbRequest.get_response("pmdata", False, True, pm_rest_api_path="http://pm-api-pro/v2/",
                                                system='LBDS', className='BSRA', source='LHC.BSRA.US45.B2',
                                                fromTimestampInNanos=t_start, durationInNanos=int(5e9))
    abort_gap_b2_df = get_lbds_parameters_from_pm_response(response_lbds_b1, parameters=['aGXpocTotalIntensity',
                                                                                         'aGXpocTotalMaxIntensity'])
    abort_gap_b2_df.columns = ['%sB2' % col for col in abort_gap_b2_df.columns]

    # Summary
    lhc_data_df = pd.concat([pm_fill_num_df, lhc_parameters_df, lhc_bm_df], axis=1)
    lhc_data_df['timestamp_blm'] = lhc_data_df.index
    lhc_data_df.index = pm_fill_num_df['pmFillNum']
    lhc_data_df.drop(columns=['pmFillNum'], inplace=True)

    abort_gap_df = pd.concat([abort_gap_b1_df, abort_gap_b2_df], axis=1)
    abort_gap_df['timestamp_abort_gap'] = abort_gap_df.index
    abort_gap_df.index = pm_fill_num_df['pmFillNum']

    lhc_context_df = pd.concat([lhc_data_df, abort_gap_df], axis=1)
    return lhc_context_df