import warnings
import numpy as np
import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery
from lhcsmapi.metadata.MappingMetadata import MappingMetadata


class RqCircuitQuery(CircuitQuery):

    def initialize_query_table(self, query_table=None):
        super().initialize_query_table(query_table)
        # nqps crate count (different across RQ circuits)
        vf_names = MappingMetadata.get_vf_for_circuit_names(self.circuit_type, self.circuit_name)
        system_vf = 'VF_{}'.format(self.circuit_name.split('.')[0])
        self.query_table.loc[self.query_table['analysis'] == system_vf, 'exp_count'] = len(vf_names)
        # diodes
        system_diode = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        magnet_names = MappingMetadata.get_magnets_for_circuit_names(self.circuit_type, self.circuit_name)
        self.query_table.loc[self.query_table['analysis'] == system_diode, 'exp_count'] = len(magnet_names)

    def update_query_table_diode_leads_pm_nxcals(self, dfss, nqps_db, analysis, db, is_ref):
        if self.query_table is None:
            return None

        signal_type = 'REF' if is_ref else 'ACT'

        # update count
        pm_count = list(nqps_db).count('PM')
        loc_mask = (self.query_table['analysis'] == analysis) & (self.query_table['db'] == 'PM') & (
                self.query_table['type'] == signal_type)
        self.query_table.loc[(self.query_table['signal'] == 'IAB.I_A') & loc_mask, ['exp_count']] = 2 * pm_count
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQD') & loc_mask, ['exp_count']] = 2 * pm_count
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQF') & loc_mask, ['exp_count']] = 2 * pm_count
        self.query_table.loc[(self.query_table['signal'] == 'U_REF_N1') & loc_mask, ['exp_count']] = 2 * pm_count

        cals_count = len(nqps_db)
        loc_mask = (self.query_table['analysis'] == analysis) & (self.query_table['db'] == 'NXCALS') & (
                self.query_table['type'] == signal_type)
        self.query_table.loc[(self.query_table['signal'] == 'I_MEAS') & loc_mask, ['exp_count']] = 2 * cals_count
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQD') & loc_mask, ['exp_count']] = 2 * cals_count
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQF') & loc_mask, ['exp_count']] = 2 * cals_count

        for index, dfs in enumerate(dfss):
            for df in dfs:
                if not df.empty:
                    col = df.columns[0].split(':')[-1]
                    loc_mask = (self.query_table['signal'] == col) & (self.query_table['analysis'] == analysis) & (
                            self.query_table['db'] == db) & (self.query_table['type'] == signal_type)
                    self.query_table.loc[loc_mask, ['act_count']] += 1

    def update_query_table_nqps_pm_nxcals(self, dfss, nqps_db, analysis, is_ref):
        if self is None:
            return None

        signal_type = 'REF' if is_ref else 'ACT'

        # update count
        pm_count = list(nqps_db).count('PM')
        loc_mask = (self.query_table['analysis'] == analysis) & (self.query_table['db'] == 'PM') & (self.query_table['type'] == signal_type)
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQD') & loc_mask, ['exp_count']] = pm_count
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQF') & loc_mask, ['exp_count']] = pm_count
        self.query_table.loc[(self.query_table['signal'] == 'U_REF_N1') & loc_mask, ['exp_count']] = 2 * pm_count

        cals_count = len(nqps_db) - pm_count
        loc_mask = (self.query_table['analysis'] == analysis) & (self.query_table['db'] == 'NXCALS') & (self.query_table['type'] == signal_type)
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQD') & loc_mask, ['exp_count']] = cals_count
        self.query_table.loc[(self.query_table['signal'] == 'U_DIODE_RQF') & loc_mask, ['exp_count']] = cals_count
        self.query_table.loc[(self.query_table['signal'] == 'U_REF_N1') & loc_mask, ['exp_count']] = 2 * cals_count

        for index, dfs in enumerate(dfss):
            db = 'PM' if nqps_db[index] == 'PM' else 'NXCALS'

            for df in dfs:
                if not df.empty:
                    col = df.columns[0].split(':')[-1]
                    loc_mask = (self.query_table['signal'] == col) & (self.query_table['analysis'] == analysis) & (
                            self.query_table['db'] == db) & (self.query_table['type'] == signal_type)
                    self.query_table.loc[loc_mask, ['act_count']] += 1

    def create_report_analysis_template(self, source_timestamp_qds_df, timestamp_fgc_rqd,
                                        i_meas_rqd_df, i_meas_rqf_df, author):
        COLUMN_NAMES = ['circuit', 'datetime_fgc', 'timestamp_iqps', 'iqps_or_nqps_trigger', 'I_MEAS_RQD_quench',
                        'I_MEAS_RQF_quench', 'magnet', 'nqps_crate', 'origin', 'dU_QS0/dt', 'QH', 'R_DIODE_LEADS_RQD',
                        'R_DIODE_LEADS_RQF', 'nqps_db', 'comments', 'performed_by', 'lhcsmapi_version']

        report_qds = pd.DataFrame(columns=COLUMN_NAMES)

        for index, row in source_timestamp_qds_df.iterrows():
            timestamp_qds = row['timestamp']

            t_quench = (timestamp_qds - timestamp_fgc_rqd) * 1e-9
            if t_quench > i_meas_rqd_df.index[-1]:
                i_quench_rqd = i_meas_rqd_df.values[-1]
                i_quench_rqf = i_meas_rqf_df.values[-1]
                warnings.warn('Quench time beyond PM buffer. Need to query NXCALS in order to get the quench current!')
            else:
                i_quench_rqd = int(i_meas_rqd_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])
                i_quench_rqf = int(i_meas_rqf_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])

            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, row['source'])
            nqps_db = 'PM' if t_quench < 2 else 'NXCALS'
            values = [self.circuit_name.replace('D', ''), Time.to_string_short(timestamp_fgc_rqd),
                      Time.to_string_short(timestamp_qds), '', i_quench_rqd, i_quench_rqf, row['source'],
                      qps_crate, '', np.nan, '', np.nan, np.nan, nqps_db, '', author, lhcsmapi.__version__]
            report_qds.loc[len(report_qds.index)] = values

        return report_qds
